﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using Microsoft.Glee.Drawing;

namespace GraphApp
{
    class GleeVisitor : IVisitor
    {
        public Graph Graph { get; private set; }
        private int id;
        private Stack<string> nodes;

        public GleeVisitor()
        {
            Graph = new Graph("");
            Graph.GraphAttr.NodeAttr.Padding = 10;
            Graph.GraphAttr.NodeAttr.Shape = Shape.Box;
            Graph.GraphAttr.NodeAttr.XRad = 4;
            Graph.GraphAttr.NodeAttr.YRad = 4;
            id = 0;
            nodes = new Stack<string>();
        }

        public void Visit(BinaryExpression expr)
        {
            AddExpression(expr);
            AddProperty("Conversion", expr.Conversion);
            AddProperty("Left", expr.Left);
            AddProperty("Right", expr.Right);
        }

        public void Visit(ConditionalExpression expr)
        {
            AddExpression(expr);
            AddProperty("Test", expr.Test);
            AddProperty("IfTrue", expr.IfTrue);
            AddProperty("IfFalse", expr.IfFalse);
        }

        public void Visit(ConstantExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.Value);
        }

        public void Visit(Expression expr)
        {
            AddExpression(expr);
        }

        public void Visit(InvocationExpression expr)
        {
            AddExpression(expr);
            AddProperty("Expression", expr.Expression);
            AddProperties("Arguments", expr.Arguments);
        }

        public void Visit(LambdaExpression expr)
        {
            AddExpression(expr);
            AddProperty("Body", expr.Body);
            AddProperties("Parameters", expr.Parameters);
        }

        public void Visit(ListInitExpression expr)
        {
            AddExpression(expr);
            AddProperty("NewExpression", expr.NewExpression);
            // expr.Initializers
        }

        public void Visit(MemberExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.Member);
            AddProperty("Expression", expr.Expression);
        }

        public void Visit(MemberInitExpression expr)
        {
            AddExpression(expr);
            AddProperty("NewExpression", expr.NewExpression);
            // expr.Bindings
        }

        public void Visit(MethodCallExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.Method);
            AddProperty("Object", expr.Object);
            AddProperties("Arguments", expr.Arguments);
        }

        public void Visit(NewArrayExpression expr)
        {
            AddExpression(expr);
            AddProperties("Expressions", expr.Expressions);
        }

        public void Visit(NewExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.Constructor);
            AddProperties("Arguments", expr.Arguments);
            // expr.Members
        }

        public void Visit(ParameterExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.Name);
        }

        public void Visit(TypeBinaryExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.TypeOperand.Name);
            AddProperty("Expression", expr.Expression);
        }

        public void Visit(UnaryExpression expr)
        {
            AddExpression(expr);
            AppendToNodeLabel(expr.Method);
            AddProperty("Operand", expr.Operand);
            // expr.IsLifted
            // expr.IsLiftedToNull
        }

        private void AddExpression(Expression expr)
        {
            string nodeId = id.ToString();
            ++id;
            Node node = Graph.AddNode(nodeId);
            node.Attr.LabelMargin = 5;
            node.Attr.Label = (expr != null) ? expr.NodeType.ToString() : "null";
            nodes.Push(nodeId);
        }

        private void AddProperty(string name, Expression property)
        {
            property.Accept(this);
            string subNodeId = nodes.Pop();
            Edge edge = Graph.AddEdge(nodes.Peek(), name, subNodeId);
            edge.Attr.Fontcolor = Color.Red;
        }

        private void AddProperties<T>(string name, ReadOnlyCollection<T> properties)
            where T : Expression
        {
            for (int i = 0; i < properties.Count; i++)
            {
                Expression property = properties[i];
                property.Accept(this);
                string subNodeId = nodes.Pop();
                Edge edge = Graph.AddEdge(nodes.Peek(), string.Format("{0}[{1}]", name, i), subNodeId);
                edge.Attr.Fontcolor = Color.Red;
            }
        }

        private void AppendToNodeLabel(object o)
        {
            if (o == null) return;
            string label = null;
            if (o is string)
            {
                label = string.Format("\"{0}\"", o);
            }
            else if (o is System.Enum)
            {
                label = string.Format("{0}.{1}", o.GetType().Name, o);
            }
            else if (o is MemberInfo)
            {
                MemberInfo member = (MemberInfo)o;
                label = string.Format("{0}.{1}", member.DeclaringType.Name, member.Name);
            }
            else
            {
                label = o.ToString();
            }
            AppendToNodeLabel(label);
        }

        private void AppendToNodeLabel(string label)
        {
            Node node = Graph.FindNode(nodes.Peek());
            StringBuilder sb = new StringBuilder(node.Attr.Label);
            sb.AppendLine();
            sb.Append("(");
            sb.Append(label);
            sb.Append(")");
            node.Attr.Label = sb.ToString();
        }
    }
}